'****************************************************************
'*  Name    : Speedometer / Revcounter / Milometer              *
'*  Author  : Bobby Garrett                                     *
'*  Notice  : Copyright (c) 2011 Bobby Garrett                  *
'*          : All Rights Reserved                               *
'*  Date    : 23/10/2011                                        *
'*  Version : 1.0                                               *
'*  Notes   : Uses capture for speedometer / Revcounter         *
'*          : TIMER0 For Milometer                              *
'*          : TIMER1 / CCP1 For Speedometer                     *
'*          : TIMER3 / CCP2 For Revcounter                      *
'*          : TIMER2 is used for general purpose timing issues  *
'*          : GLCD for output                                   *
'*          :                                                   *
'****************************************************************  

   Device 18F452
   Config_Start
   OSC = HS ; HS
   OSCS = OFF ; Disabled
   PWRT = On ; Enabled
   BOR = On ; Enabled
   BORV = 45 ; 4.5V
   WDT = OFF ; Disabled
   WDTPS = 1 ; 1:1
   CCP2MUX = OFF ; Disable (RB3)
   STVR = On ; Enabled
   LVP = OFF ; Disabled
   Debug = OFF ; Disabled
   CP0 = OFF ; Disabled
   CP1 = OFF ; Disabled
   CP2 = OFF ; Disabled
   CP3 = OFF ; Disabled
   CPB = OFF ; Disabled
   CPD = OFF ; Disabled
   WRT0 = OFF ; Disabled
   WRT1 = OFF ; Disabled
   WRT2 = OFF ; Disabled
   WRT3 = OFF ; Disabled
   WRTB = OFF ; Disabled
   WRTC = OFF ; Disabled
   WRTD = OFF ; Disabled
   EBTR0 = OFF ; Disabled
   EBTR1 = OFF ; Disabled
   EBTR2 = OFF ; Disabled
   EBTR3 = OFF ; Disabled
   EBTRB = OFF ; Disabled
Config_End
 	     
        Xtal = 40                   'Define clock
        All_Digital = True          'Set all I/O to digital	 
 	     
'****************************************************************          
'* DISPLAY                                                      *
'****************************************************************           
        LCD_DTPort = PORTD
		LCD_RSPin = PORTC.1
		LCD_ENPin = PORTE.0
		LCD_RWPin = PORTC.0
        LCD_CS1Pin = PORTE.1
		LCD_CS2Pin = PORTE.2
		
		LCD_Type = GRAPHIC
		Internal_Font = On            
       
        GLCD_EXTERNAL_PRINT = PPRINT	'Point to PPRINT.INC file

        Clear                       'Clear all RAM before we start
    
' Variables And declares for Speed capture   
	    Dim SPEED_PERIOD As Dword     'Time interval between pulses
        Dim SPEED_FREQ As Float
        Dim SPEED As Float 	
        Dim SPEED_ROLLOVER As Dword    
        Dim SPEED_MULTIPLIER As Float  
        
'Capture settings - used for speed                        
	   Symbol CCPR1 = CCPR1L.Word	'Create a 16-bit variable out of CCPRL1
	
	   CCP1CON = %00000110		    'Enable the CCP1 capture, every 4th rising edge	
'0100 = Capture mode, every falling edge
'0101 = Capture mode, every rising edge
'0110 = Capture mode, every 4th rising edge
'0111 = Capture mode, every 16th rising edge    	
	
        Symbol SPEED_CAPTURE = PIR1.2	'CCP1 capture flag
'Timer 0 settings - used for distance    	
        Symbol TIMER0 = TMR0L.Word	   'Create a 16-bit variable out of TMR0L
        T0CON = %10100001		       'TMR0 as counter, low to high, PSA 1:4 
'Timer 1 settings - used for speed      	
        Symbol SPEED_TIMER = TMR1L.Word	   'Create a 16-bit variable out of TMR1L
        T1CON = %00000001		       'TMR1 prescale=1:1, and turn it on (100nS per count)
'bits 5-4
'00 = 1:1
'01 = 1:2 
'10 = 1:4
'11 = 1:8       
        
        Symbol SPEED_TMR_OVERFLOW	= PIR1.0	'Timer1 overflow flag          

'Variables and declares for RPM capture      	
        Dim RPM_PERIOD As Dword	
        Dim RPM_FREQ As Float	
        Dim RPM_ROLLOVER As Dword
        Dim RPM As Dword    
    
        Symbol CCPR2 = CCPR2L.Word		 'Create a 16-bit variable out of CCPRL2
        Symbol RPM_TIMER = TMR3L.Word	 'Create a 16-bit variable out of TMR3L
        Symbol RPM_CAPTURE = PIR2.0		 'CCP2 capture flag
        Symbol RPM_TMR_OVERFLOW = PIR2.1 'Timer3 overflow flag   
    
        CCP2CON = %00000110		         'Enable the CCP2 capture, every rising edge
'0100 = Capture mode, every falling edge
'0101 = Capture mode, every rising edge
'0110 = Capture mode, every 4th rising edge
'0111 = Capture mode, every 16th rising edge    	
	
        T3CON = %00001001		        'TMR1 prescale=1:8, and turn it on (8uS per count)		
'bits 5-4                               'Set Timer 1 for CCP1 and Timer 3 for CCP2
'00 = 1:1
'01 = 1:2 
'10 = 1:4
'11 = 1:8                                       
     
        T2CON = %00000111   'Turn timer on / PSA of 16:1
        PR2 = 250           '250 and PSA of 16 gives 400uS interrupt with 40mhz xtal
	   
' Variables for speed calculations
        Dim WHEEL_CIRC As Word     'Wheel circumference in millimeters
        Dim NR_OF_BOLTS As Byte    'Number of disk bolts or other per revolution
        Dim PITCH As Float         'WHEEL_CIRC / NR_OF_BOLTS for speed and dist calcs

'Variables for Digital display
        Dim SPEED_INT As Word       'Used to convert float to integer
        Dim SPEED_UNITS As Byte     'Used for digital output
        Dim SPEED_TENS As Byte      'Used For digital Output
        Dim SPEED_HUNDREDS As Byte  'Used For digital Output

        Dim UPDATE_REQ As Byte

        Dim GP_TIMER_1 As Word

'****************************************************************  
' User defined configuration settings                           *
'****************************************************************     
        WHEEL_CIRC = 1995           'Enter wheel circumference in MM
        NR_OF_BOLTS = 6             'Enter number of pulses per rev (Disc bolts)

'**************************************************************** 
' End of user defined settings                                  *
'**************************************************************** 

'PORTS setup
        TRISA.4 = 1                 'Set PortA.4 as input for counter
        TRISC.2 = 1                 'SET CCP1 as input (Speed)
        TRISB.3 = 1                 'Set CCP2 as input (RPM)                 

'INTERRUPT setup  
'Used for speed  
'   PIE1.2 = 1                      'CCP1 Interrupt enable
'   IPR1.2 = 1                      'CCP1 Interrupt = High priority
'   PIE1.0 = 1                      'Timer 1 overflow Interrupt Enable
'   IPR1.0 = 1                      'Timer 1 overflow Int = High priority

'Used for RPM    
'   PIE2.0 = 1                      'CCP2 Interrupt enable
'   IPR2.0 = 0                      'CCP2 Interrupt = Low priority
'   PIE2.1 = 1                      'Timer 3 overflow Interrupt Enable
'   IPR2.1 = 0                      'Timer 3 overflow Int = Low priority

'Used for gereral purpose timer functions
'   PIE1.1 = 1                      'Timer2 = PR2 interrupt enable
'   IPR1.1 = 0                      'Timer2 = PR2 interrupt = LOW PRIORITY
        
        PIE1 = %00000111
        IPR1 = %00000101  
        PIE2 = %00000011
        IPR2 = %00000000    
 
'Enable priority interrupts    
        RCON.7 = 1 
 
'****************************************************************
'* Details of speed formula calculations                        *
'****************************************************************    

'Calculate the speed formula
'KPH - SPEED = FREQUENCY * PITCH * 3600
'              ------------------------
'              1,000,000 (MM in a Kilometer)

'MPH - SPEED = FREQUENCY * PITCH * 3600
'              ------------------------
'              1,609,344 (MM in a mile)
  
'Calculate distance between pulses    
        PITCH = WHEEL_CIRC / NR_OF_BOLTS 'MM per pulse for speed calculation
        
'Calculate multiplication factor for KPH 
   '     If MPH_KPH = KPH Then       		
   '     SPEED_MULTIPLIER = PITCH * 3.6
   '     SPEED_MULTIPLIER = SPEED_MULTIPLIER / 1000

'Calculate multiplication factor for MPH         
        SPEED_MULTIPLIER = PITCH * 2.237
        SPEED_MULTIPLIER = SPEED_MULTIPLIER / 1000

'****************************************************************
'* End of details of SPEED formula calculations                 *
'****************************************************************
    
'****************************************************************
'* Interrupt service vector initialization                      *
'****************************************************************

On_Hardware_Interrupt GoTo HIGH_INTERRUPT_ROUTINE  ' Point the High Priority interrupt handler to the subroutine  
On_Low_Interrupt GoTo LOW_INTERRUPT_ROUTINE        ' Point the Low Priority interrupt handler to the subroutine
                                                          
        GoTo OVER_INT  
        
'****************************************************************       
'* High Interrupt is used for Speed capture and Low Interrupt   *
'* is used for RPM capture                                      *
'****************************************************************
'****************************************************************       
'* High Interrupt Service Routine                               *
'****************************************************************

High_Int_Sub_Start
HIGH_INTERRUPT_ROUTINE:

        Context Save   
        If SPEED_TMR_OVERFLOW = 1 Then GoTo SPEED_ROLL_OVER       
        If SPEED_CAPTURE = 1 Then GoTo SPEED_VAL  'Calc Speed value    

        Context Restore                 'Exit ISR and re-enable interrupts      

SPEED_ROLL_OVER:        
        Inc SPEED_ROLLOVER
        If SPEED_ROLLOVER > 200 Then GoTo SPEED_ZERO  '200 equates to approx 4 kph

    	SPEED_TMR_OVERFLOW = 0			'Clear Timer1 overflow flag

        If SPEED_CAPTURE = 1 Then GoTo SPEED_VAL  'Calc Speed value
        
        Context Restore                 'Exit ISR and re-enable interrupts      
        
SPEED_VAL:
        SPEED_PERIOD = SPEED_ROLLOVER * 65535
    	SPEED_PERIOD = SPEED_PERIOD + CCPR1       'Store the captured value
    	
        SPEED_FREQ = 10000000.0 / (SPEED_PERIOD / 4) 
        
        SPEED = SPEED_FREQ * SPEED_MULTIPLIER
             
      	Clear SPEED_TIMER			    'Clear Timer1
      	SPEED_ROLLOVER = 0
      	SPEED_PERIOD = 0             
       	SPEED_FREQ = 0    
    	SPEED_CAPTURE = 0			    'Clear CCP1 capture flag
    	SPEED_TMR_OVERFLOW = 0		    'Clear Timer1 overflow flag	
	
        UPDATE_REQ = 1                  'Set flag to call display update outside
                                        'of interrupt routine
                                        
        Context Restore                 'Exit ISR and re-enable interrupts   

SPEED_ZERO:
        CCP1CON = %00000000

        SPEED = 0                    
                
      	Clear SPEED_TIMER			    'Clear Timer1
      	SPEED_ROLLOVER = 0      	
      	SPEED_PERIOD = 0       
      	SPEED_FREQ = 0    
    	SPEED_CAPTURE = 0			    'Clear CCP1 capture flag
    	SPEED_TMR_OVERFLOW = 0		    'Clear Timer1 overflow flag	
	
        UPDATE_REQ = 1                  'Set flag to call display update outside
                                        'of interrupt routine

        CCP1CON = %00000110
   
        Context Restore                 'Exit ISR and re-enable interrupts  

High_Int_Sub_End

'****************************************************************
'* End of High Interrupt routine                                *
'****************************************************************        
'****************************************************************       
'* Low Interrupt Service Routine                                *
'****************************************************************

Low_Int_Sub_Start                                                   
LOW_INTERRUPT_ROUTINE:
        Context Save 
     
        If RPM_TMR_OVERFLOW = 1 Then GoTo RPM_ROLL_OVER     
        If RPM_CAPTURE = 1 Then GoTo RPM_VAL 'Calc RPM value   

        If PIR1.1 = 1 Then GoTo TIMED_FUNCTIONS

        Context Restore                 'Exit ISR and re-enable interrupts      

RPM_ROLL_OVER:        
        Inc RPM_ROLLOVER
        If RPM_ROLLOVER > 100 Then GoTo RPM_ZERO  '100 equates to approx 400 rpm

    	RPM_TMR_OVERFLOW = 0			 'Clear Timer3 overflow flag

        If RPM_CAPTURE = 1 Then GoTo RPM_VAL   'Calc RPM value
        
        Context Restore                 'Exit ISR and re-enable interrupts      
        
RPM_VAL:

        RPM_PERIOD = RPM_ROLLOVER * 65535
    	RPM_PERIOD = RPM_PERIOD + CCPR2     'Store the captured value
        RPM_FREQ = 10000.0 / (RPM_PERIOD / 4)

        RPM = RPM_FREQ * 6000           'Maths is done this way so RPM only
        RPM = RPM * 10                  'increments in 10 RPM intervals  
                                        'for 100 rpm steps use 600 / 100.
               
      	Clear RPM_TIMER			        'Clear Timer3
      	RPM_ROLLOVER = 0
      	RPM_PERIOD = 0           
       	RPM_FREQ = 0    
    	RPM_CAPTURE = 0				    'Clear CCP2 capture flag
    	RPM_TMR_OVERFLOW = 0		    'Clear Timer3 overflow flag	
	
        UPDATE_REQ = 1                  'Set flag to call display update outside
                                        'of interrupt routine
                                        
        Context Restore                 'Exit ISR and re-enable interrupts   

RPM_ZERO:
        CCP2CON = %00000000

        RPM = 0                    
                
      	Clear RPM_TIMER			        'Clear Timer3
      	RPM_ROLLOVER = 0      	
      	RPM_PERIOD = 0       
      	RPM_FREQ = 0   
       	RPM_CAPTURE = 0				    'Clear CCP2 capture flag
    	RPM_TMR_OVERFLOW = 0		    'Clear Timer3 overflow flag	
	
        UPDATE_REQ = 1                  'Set flag to call display update outside
                                        'of interrupt routine

        CCP2CON = %00000110
   
        Context Restore                 'Exit ISR and re-enable interrupts  
        
TIMED_FUNCTIONS: 
'This routine is called approx every 400uS 
'PR2 = 250 and PSA of 16 gives 400uS interrupt with 40mhz xtal 
'Use it to increment counters and compare them for various tasks.
'A count of
'25 =   10mS   (100 / sec) intervals. 
'250 =  100mS  (10 / sec) intervals. 
'1250 = 500mS  (2 / sec) intervals. 
'2500 = 1000mS (1 / sec) intervals.     

        Inc GP_TIMER_1
        
        PIR1.1 = 0

        Context Restore                 'Exit ISR and re-enable interrupts  

Low_Int_Sub_End

'****************************************************************
'* End of Interrupt routines                                    *
'****************************************************************        

OVER_INT:     

'***********************************************
'* Start of Main Code                          *
'***********************************************
START:
        DelayMS 100                     'Wait for the PICmicro to stabilise
        Cls                             'Clear the LCD
        
        Print Font Arial____8o25
        Print At 38,87, "Mph"
   
        GoSub UPDATE
         
        INTCON = %11000000      'Turn interrupts on    
            
'***********************************************
'* Main program LOOP                           *
'***********************************************
MAIN_LOOP:

        If UPDATE_REQ = 1 Then  GoSub UPDATE 
 	
        GoTo MAIN_LOOP                  		

'This routine is called when there is a reading to be displayed.
'It is also called on timer 1 overflow if there is no input.
'This occurs approximately every 1/2 second, other code can go here.
'Be aware that at high RPM it will be called more so do not use it for timing. 
'Do not alter timer 1 settings to call this routine more frequently as it will
'mess up the low end of the revcounter.
UPDATE:
        GoSub DISPLAY_RPM               'RPM display
        GoSub DISPLAY_DIGITAL           'Digital display
        UPDATE_REQ = 0
        Return

DISPLAY_RPM:
         Print Font Arial____8o25   'Small fonts
         Print At 4,60,Dec RPM, " Rpm   "
         Return    
       
DISPLAY_DIGITAL:    
        Print Font Arial____24     'Large fonts

'Turn Float to integer as Dig command does not work with Float
        SPEED_INT = SPEED

'Split the value into individual numbers so that I can right justify the output 
        SPEED_HUNDREDS = Dig SPEED_INT,2
        SPEED_TENS = Dig SPEED_INT,1
        SPEED_UNITS = Dig SPEED_INT,0

        If SPEED_INT > 99 Then
        Print At 22,28,Dec SPEED_HUNDREDS
        Else
        Print At 22,28, " "
        EndIf
        If SPEED_INT > 9 Then
        Print At 22,48,Dec SPEED_TENS
        Else
        Print At 22,48, " "
        EndIf
        Print At 22,68,Dec SPEED_UNITS
       
        Return
    
Include "ARIAL_8_25.INC"		'DATA FOR ARIAL_8_25 FONT
Include "ARIAL_24.INC"			'DATA FOR ARIAL_24 FONT 
